#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#ifdef USE_ERR
#include <gsl/gsl_statistics_long_double.h>
#endif

#ifdef _OPENMP
#include <omp.h>
#endif

#ifdef USE_RESAMPLE
#include <hdf5.h>
#include <hdf5_hl.h>
#endif

#include "common.h"

#ifdef USE_ERR
long double cal_amplitude(long double *date_array, long double *norm_mag, long double *error_array, int total_number, long double trial_freq, long double variance, long double degree, long double factor1, long double factor2, long double error_sum) ;
#else
long double cal_amplitude(long double *date_array, long double *norm_mag, long double *error_array, int total_number, long double trial_freq, long double variance, long double degree, long double factor1, long double factor2) ;
#endif

#ifdef USE_RESAMPLE
#ifdef USE_ERR
int jackknife_sample(long double *date_array, long double *mag_array, long double *error_array, int total_number, int out_ind, long double error_sum, long double *jackknife_date_array, long double *jackknife_norm_mag, long double *jackknife_error_array, long double *new_variance, long double *sample_error_sum);
#else
int jackknife_sample(long double *date_array, long double *mag_array, long double *error_array, int total_number, int out_ind, long double *jackknife_date_array, long double *jackknife_norm_mag, long double *jackknife_error_array, long double *new_variance);
#endif
#endif

typedef struct {
		long double r;
		long double i;
} complex;

complex Csqrt(complex z);
complex Cdiv(complex a, complex b);
complex Cmul(complex a, complex b);
complex Cconj(complex z);

int main(int argc, char **argv)
{
	int total_number, id_n, num_loop, i;
	FILE *input, *output;
	char *result_file=".multi.pdg";
	long double *date_array, *mag_array, *error_array, *norm_mag;
	long double min_freq,max_freq,trial_freq,freq_interval;
#ifdef USE_ERR
	long double error_sum;
#ifdef USE_RESAMPLE
	long double sample_error_sum;
#endif
#endif
#ifdef USE_RESAMPLE
	int total_number_minus_1;
	char hdf5_resample_file[256];
	char hdf5_dset_name[256];
	hid_t hdf5_file_id;
	hsize_t hdf5_dims[1];
	herr_t hdf5_status;
	long double *sample_date_array, *sample_norm_mag, *sample_error_array;
	double *F_sample_array;
	long double sample_variance, sample_factor1;
#endif
	long double mean_mag,temp,temp_freq,F;
	long double variance, degree, factor1, factor2;
	long double min_period_factor=1.00, max_period_factor=1.00;
	int opt_tag[5] = {0, 0, 0, 0, 0};
	long double opt_value[5];

#ifdef _OPENMP
        int num_cpus;
        long double *P_array;
        long double *f_array;
        long double *F_array;
        num_cpus = omp_get_num_procs();
        printf("# OpenMP is working with %d CPUs\n", num_cpus);
        omp_set_num_threads(num_cpus);
#endif

	if(argc < 3)
	{
		usage_option();
		return 1;
	}

	total_number = atoi(argv[2]);
	#ifdef USE_RESAMPLE
	total_number_minus_1 = total_number - 1;
	#endif

	date_array = calloc(total_number, sizeof(long double));
	if(date_array == NULL) {fprintf(stderr, "Memory problem: date_array\n"); return 1;}
	mag_array = calloc(total_number, sizeof(long double));
	if(mag_array == NULL) {fprintf(stderr, "Memory problem: mag_array\n"); return 1;}
	error_array = calloc(total_number, sizeof(long double));
	if(error_array == NULL) {fprintf(stderr, "Memory problem: error_array\n"); return 1;}
	norm_mag = calloc(total_number, sizeof(long double));
	if(norm_mag == NULL) {fprintf(stderr, "Memory problem: norm_mag\n"); return 1;}

	if( parse_option(argc, argv, &opt_tag[0], &opt_value[0]) != 0 ) {
		usage_option();
		return 1;
	}
	if( opt_tag[3] != 0 ) {
		min_period_factor = opt_value[3];
	}
	if( opt_tag[4] != 0 ) {
		max_period_factor = opt_value[4];
	}

	input = fopen(argv[1], "r");
	#ifdef USE_RESAMPLE
	sprintf(hdf5_resample_file, "%s.multi.sample.h5", argv[1]);
	#endif
	result_file = strcat(argv[1],result_file);

	
/* Read data file with three cols: date, mag., and mag. error */
	
	for(id_n = 0; id_n < total_number; ++id_n) {
			fscanf(input,"%Lf %Lf %Lf",(date_array+id_n),(mag_array+id_n),(error_array+id_n));
	}

/* Mean and Variance of Mag. */

	for(id_n = 0; id_n < total_number; ++id_n) {
#ifdef USE_ERR
                        if((*(error_array+id_n)) != 0.0 )
                        {
                                *(error_array + id_n) = 1.0 / ((*(error_array+id_n)) * (*(error_array+id_n)));
                        }
                        else
                        {
                                        (*(error_array+id_n)) = 0.0004; // default error value for non-sense 0.
                                        *(error_array + id_n) = 1.0 / ((*(error_array+id_n)) * (*(error_array+id_n)));
                        }
#else
			*(error_array + id_n) = (*(error_array+id_n)) * (*(error_array+id_n));
#endif
	}

	mean_mag = 0.0;
	variance = 0.0;
#ifdef USE_ERR
	mean_mag = gsl_stats_long_double_wmean(error_array, 1, mag_array, 1, total_number);
	error_sum = 0.0;
        for(id_n = 0; id_n < total_number; ++id_n) {
                        *(norm_mag+id_n) = (*(mag_array + id_n) - mean_mag);
                        error_sum = error_sum + *(error_array + id_n);
        }
        variance = gsl_stats_long_double_wvariance_m(error_array, 1, mag_array, 1, total_number, mean_mag);
        variance = variance * (total_number - 1.0);
#else
	for(id_n = 0; id_n < total_number; ++id_n) {
			mean_mag = mean_mag + *(mag_array + id_n);
	}
	mean_mag = mean_mag / total_number;
	for(id_n = 0; id_n < total_number; ++id_n) {
			*(norm_mag+id_n) = (*(mag_array + id_n) - mean_mag);
	}
	for(id_n = 0; id_n < total_number; ++id_n) {
		variance = variance + (*(norm_mag+id_n)) * (*(norm_mag+id_n)) ;
	}
#endif

/* Trial frequency setting */

	temp = (*(date_array+total_number-1)) - (*date_array);
	min_freq = 2.0 * M_PI / temp;
	max_freq = min_freq;
	for(id_n = 0; id_n < (total_number-1); ++id_n) {
			temp_freq = M_PI / (*(date_array+id_n+1) - *(date_array+id_n));
			if( temp_freq > max_freq ) {
				max_freq = temp_freq;
			}
	}
	if( max_freq > (2.0 * M_PI / 0.000001) )
	{
		max_freq = (2.0 * M_PI / 0.000001);
	}

	printf("# Min f. %Lf Max f. %Lf Min P. %Lf Max P. %Lf\n", min_freq, max_freq, (2.0 * M_PI / max_freq), (2.0 * M_PI / min_freq));

	/* User-given limits */
	if(opt_tag[0] != 0) { // minimum period
		max_freq = (2.0 * M_PI / opt_value[0]);
	}
	if(opt_tag[1] != 0) { // maximum period
		min_freq = (2.0 * M_PI / opt_value[1]);
	}
	if( opt_tag[3] != 0 ) {
		max_freq = max_freq / min_period_factor;
	}
	if( opt_tag[4] != 0 ) {
		min_freq = min_freq / max_period_factor;
	}
	if( (opt_tag[0] != 0) || (opt_tag[1] != 0) || (opt_tag[3] != 0) || (opt_tag[4] != 0) ) {
		printf("... using Min f. %Lf Max f. %Lf Min P. %Lf Max P. %Lf\n",min_freq,max_freq,(2.0 * M_PI / max_freq),(2.0 * M_PI / min_freq));
	}
	if(opt_tag[2] == 0) {
		opt_value[2] = DEFAULT_dP_over_P;
	}
	if(opt_tag[2] == -1) { // constant step size
		opt_value[2] = fabsl(opt_value[2]);
	}
	/* Validation of trial periods range */
	if(max_freq < min_freq) {
		printf("... the period range does not work.\n");
		usage_option();
		return 1;
	}
		
	
	/* 'degree' parameter determines the degree of harmonic terms. You can change it. */
	degree = USE_DEGREE;
	factor2 = degree * 2.0;
	factor1 = ((long double) total_number) - factor2 - 1.0;
	#ifdef USE_RESAMPLE
	sample_factor1 = factor1 - 1.0;
	#endif

	output = fopen(result_file, "w");

	trial_freq = min_freq;
        #ifdef _OPENMP // OpenMP parallelization.
        num_loop = 0;
        while (trial_freq <= max_freq) {
                        num_loop += 1;
			if(opt_tag[2] < 0) { // dP = constant.
	                        freq_interval = trial_freq * trial_freq * opt_value[2] * 0.50 * M_1_PI;
			} else { // dP/P = constant.
        	                freq_interval = trial_freq * opt_value[2];
			}
                        trial_freq = trial_freq + freq_interval;
        }
        P_array = calloc(num_loop, sizeof(long double));
	if(P_array == NULL) {fprintf(stderr, "Memory problem: P_array\n"); return 1;}
        f_array = calloc(num_loop, sizeof(long double));
	if(f_array == NULL) {fprintf(stderr, "Memory problem: f_array\n"); return 1;}
        F_array = calloc(num_loop, sizeof(long double));
	if(F_array == NULL) {fprintf(stderr, "Memory problem: F_array\n"); return 1;}
        trial_freq = min_freq;
        for(i=0; i < num_loop; i++) {
                        *(P_array + i) = (2.0 * M_PI) / trial_freq;
                        *(f_array + i) = trial_freq;
			if(opt_tag[2] < 0) { // dP = constant.
	                        freq_interval = trial_freq * trial_freq * opt_value[2] * 0.50 * M_1_PI;
			} else { // dP/P = constant.
        	                freq_interval = trial_freq * opt_value[2];
			}
                        trial_freq = trial_freq + freq_interval;
        }
	#else // OpenMP parallelization.
	#ifdef USE_RESAMPLE
        num_loop = 0;
        while (trial_freq <= max_freq) {
                        num_loop += 1;
			if(opt_tag[2] < 0) { // dP = constant.
	                        freq_interval = trial_freq * trial_freq * opt_value[2] * 0.50 * M_1_PI;
			} else { // dP/P = constant.
        	                freq_interval = trial_freq * opt_value[2];
			}
                        trial_freq = trial_freq + freq_interval;
        }
	#endif
        #endif // OpenMP parallelization.
	#ifdef USE_RESAMPLE
	hdf5_dims[0] = num_loop;
        F_sample_array = calloc(num_loop, sizeof(double));
	if(F_sample_array == NULL) {fprintf(stderr, "Memory problem: F_sample_array\n"); return 1;}
	hdf5_file_id = H5Fcreate(hdf5_resample_file, H5F_ACC_TRUNC, H5P_DEFAULT, H5P_DEFAULT);
	#endif

	#ifdef _OPENMP // OpenMP parallelization.
	// Estimation by using the original data.
#pragma omp parallel default(shared)
#pragma omp for schedule(dynamic) private(id_n,F,trial_freq)
        for(id_n=0; id_n < num_loop; id_n++) {
                        trial_freq = *(f_array + id_n);
#ifdef USE_ERR
			F = cal_amplitude(date_array, norm_mag, error_array, total_number, trial_freq, variance, degree, factor1, factor2, error_sum);
#else
                        F = cal_amplitude(date_array, norm_mag, error_array, total_number, trial_freq, variance, degree, factor1, factor2);
#endif
                        *(F_array + id_n) = F;
        }
	#ifdef USE_RESAMPLE // Resampling starts when OpenMP is on.
	sample_date_array = calloc(total_number_minus_1, sizeof(long double));
	if(sample_date_array == NULL) {fprintf(stderr, "Memory problem: sample_date_array\n"); return 1;}
	sample_norm_mag = calloc(total_number_minus_1, sizeof(long double));
	if(sample_norm_mag == NULL) {fprintf(stderr, "Memory problem: sample_norm_mag\n"); return 1;}
	sample_error_array = calloc(total_number_minus_1, sizeof(long double));
	if(sample_error_array == NULL) {fprintf(stderr, "Memory problem: sample_error_array\n"); return 1;}
	for(i=0; i < total_number; ++i) {
		#ifdef USE_ERR
		jackknife_sample(date_array, mag_array, error_array, total_number, i, error_sum, \
		sample_date_array, sample_norm_mag, sample_error_array, &sample_variance, &sample_error_sum);
#pragma omp parallel default(shared)
#pragma omp for schedule(dynamic) private(id_n,F,trial_freq)
		for(id_n=0; id_n < num_loop; id_n++) {
                        trial_freq = *(f_array + id_n);
			F = cal_amplitude(sample_date_array, sample_norm_mag, sample_error_array, total_number_minus_1, \
			trial_freq, sample_variance, degree, sample_factor1, factor2, sample_error_sum);
                        *(F_sample_array + id_n) = F; // Jackknife replication.
		}
		#else
		jackknife_sample(date_array, mag_array, error_array, total_number, i, \
		sample_date_array, sample_norm_mag, sample_error_array, &sample_variance);
#pragma omp parallel default(shared)
#pragma omp for schedule(dynamic) private(id_n,F,trial_freq)
		for(id_n=0; id_n < num_loop; id_n++) {
                        trial_freq = *(f_array + id_n);
			F = cal_amplitude(sample_date_array, sample_norm_mag, sample_error_array, total_number_minus_1, \
			trial_freq, sample_variance, degree, sample_factor1, factor2);
                        *(F_sample_array + id_n) = F; // Jackknife replication.
		}
		#endif
		sprintf(hdf5_dset_name, "/sample%d", i+1);
		hdf5_status = H5LTmake_dataset(hdf5_file_id, hdf5_dset_name, 1, hdf5_dims, H5T_NATIVE_DOUBLE, F_sample_array);
	}
	free(sample_date_array);
	free(sample_norm_mag);
	free(sample_error_array);
	#endif // Resampling starts when OpenMP is on.
        #else // OpenMP parallelization.
	// Estimation by using the original data.
        trial_freq = min_freq;
	while (trial_freq <= max_freq) {
#ifdef USE_ERR
			F = cal_amplitude(date_array, norm_mag, error_array, total_number, trial_freq, variance, degree, factor1, factor2, error_sum);
#else
			F = cal_amplitude(date_array, norm_mag, error_array, total_number, trial_freq, variance, degree, factor1, factor2);
#endif
			temp = (2.0 * M_PI) / trial_freq;
			fprintf(output,"%.6Lf %.6Lf\n",temp, F);
			if(opt_tag[2] < 0) { // dP = constant.
	                        freq_interval = trial_freq * trial_freq * opt_value[2] * 0.50 * M_1_PI;
			} else { // dP/P = constant.
        	                freq_interval = trial_freq * opt_value[2];
			}
			trial_freq = trial_freq + freq_interval;
	}
	#ifdef USE_RESAMPLE // Resampling starts when OpenMP is off.
	sample_date_array = calloc(total_number_minus_1, sizeof(long double));
	if(sample_date_array == NULL) {fprintf(stderr, "Memory problem: sample_date_array\n"); return 1;}
	sample_norm_mag = calloc(total_number_minus_1, sizeof(long double));
	if(sample_norm_mag == NULL) {fprintf(stderr, "Memory problem: sample_norm_mag\n"); return 1;}
	sample_error_array = calloc(total_number_minus_1, sizeof(long double));
	if(sample_error_array == NULL) {fprintf(stderr, "Memory problem: sample_error_array\n"); return 1;}
	for(i=0; i < total_number; ++i) {
		#ifdef USE_ERR
		jackknife_sample(date_array, mag_array, error_array, total_number, i, error_sum, \
		sample_date_array, sample_norm_mag, sample_error_array, &sample_variance, &sample_error_sum);
		trial_freq = min_freq;
		id_n = 0;
		while (trial_freq <= max_freq) {
			F = cal_amplitude(sample_date_array, sample_norm_mag, sample_error_array, total_number_minus_1, \
			trial_freq, sample_variance, degree, sample_factor1, factor2, sample_error_sum);
                        *(F_sample_array + id_n) = F; // Jackknife replication.
			if(opt_tag[2] < 0) { // dP = constant.
	                        freq_interval = trial_freq * trial_freq * opt_value[2] * 0.50 * M_1_PI;
			} else { // dP/P = constant.
        	                freq_interval = trial_freq * opt_value[2];
			}
			trial_freq = trial_freq + freq_interval;
			id_n = id_n + 1;
		}
		#else
		jackknife_sample(date_array, mag_array, error_array, total_number, i, \
		sample_date_array, sample_norm_mag, sample_error_array, &sample_variance);
		trial_freq = min_freq;
		id_n = 0;
		while (trial_freq <= max_freq) {
			F = cal_amplitude(sample_date_array, sample_norm_mag, sample_error_array, total_number_minus_1, \
			trial_freq, sample_variance, degree, sample_factor1, factor2);
                        *(F_sample_array + id_n) = F; // Jackknife replication.
			if(opt_tag[2] < 0) { // dP = constant.
	                        freq_interval = trial_freq * trial_freq * opt_value[2] * 0.50 * M_1_PI;
			} else { // dP/P = constant.
        	                freq_interval = trial_freq * opt_value[2];
			}
			trial_freq = trial_freq + freq_interval;
			id_n = id_n + 1;
		}
		#endif
		sprintf(hdf5_dset_name, "/sample%d", i+1);
		hdf5_status = H5LTmake_dataset(hdf5_file_id, hdf5_dset_name, 1, hdf5_dims, H5T_NATIVE_DOUBLE, F_sample_array);
	}
	free(sample_date_array);
	free(sample_norm_mag);
	free(sample_error_array);
	#endif // Resampling starts when OpenMP is off.
        #endif // OpenMP parallelization.
#ifdef _OPENMP
        for(id_n=0; id_n < num_loop; id_n++) {
                fprintf(output,"%.6Lf %.6Lf\n",*(P_array + id_n), *(F_array + id_n));
        }
        free (P_array);
        free (f_array);
        free (F_array);
#endif

	fclose(input);
	fclose(output);

	free (date_array);
	free (mag_array);
	free (error_array);
	free (norm_mag);

	#ifdef USE_RESAMPLE
	hdf5_status = H5Fclose(hdf5_file_id);
	#endif
	
	return 0;
}

#ifdef USE_ERR
long double cal_amplitude(long double *date_array, long double *norm_mag, long double *error_array, int total_number, long double trial_freq, long double variance, long double degree, long double factor1, long double factor2, long double error_sum)
{
        complex *c, *z, *zn, *p, *cf, temp, temp1, temp2;
        long double F,sav;
        int n, k;
        complex a, sn, sc, sad, san;

        c = calloc((factor2+1), sizeof(complex));
	if(c == NULL) {fprintf(stderr, "Memory problem: c\n"); abort();}
        z = calloc(total_number, sizeof(complex));
	if(z == NULL) {fprintf(stderr, "Memory problem: z\n"); abort();}
        zn = calloc(total_number, sizeof(complex));
	if(zn == NULL) {fprintf(stderr, "Memory problem: zn\n"); abort();}
        p = calloc(total_number, sizeof(complex));
	if(p == NULL) {fprintf(stderr, "Memory problem: p\n"); abort();}
        cf = calloc(total_number, sizeof(complex));
	if(cf == NULL) {fprintf(stderr, "Memory problem: cf\n"); abort();}
        
        F = 0.0;
        a.r = 0.0;
        a.i = 0.0;

        for( n=0; n < total_number; n++) {
                        sav = fmod((trial_freq * (*(date_array+n))) , (2.0 * M_PI) );
                (*(z+n)).r = cos(sav);
                (*(z+n)).i = sin(sav);
                        (*(zn+n)).r = (Cconj(*(z+n))).r;
                        (*(zn+n)).i = (Cconj(*(z+n))).i;
                        (*(p+n)).r = (*(zn+n)).r;
                        (*(p+n)).i = (*(zn+n)).i;
                        (*(cf+n)).r = (*(norm_mag+n)) * cos(sav * degree);
                        (*(cf+n)).i = (*(norm_mag+n)) * sin(sav * degree);
        }

        for( n=0; n <= factor2; n++) {

                        sn.r = 0.0;
                        sn.i = 0.0;
                        sc.r = 0.0;
                        sc.i = 0.0;
                        sad.r = 0.0;
                        sad.i = 0.0;
                        san.r = 0.0;
                        san.i = 0.0;

                        for( k = 0; k < total_number; k++) {
                                
                                temp1 = Cmul((*(z+k)), (*(p+k)));
                                temp2 = Cmul((*(zn+k)), Cconj(*(p+k)) );
                                temp2 = Cmul(a , temp2);
                                (*(p+k)).r = temp1.r - temp2.r;
                                (*(p+k)).i = temp1.i - temp2.i;

                                temp1 = Cmul( *(p+k) , Cconj(*(p+k)) );
                                sn.r = sn.r + (*(error_array+k)) * temp1.r;
                                sn.i = sn.i + (*(error_array+k)) * temp1.i;
                                temp2 = Cmul((*(cf+k)), Cconj(*(p+k)) );
                                sc.r = sc.r + (*(error_array+k)) * temp2.r;
                                sc.i = sc.i + (*(error_array+k)) * temp2.i;

                                (*(zn+k)) = Cmul((*(zn+k)), (*(z+k)));
                                
                                
                                temp1 = Cmul((*(zn+k)), Cconj(*(p+k)));
                                temp1 = Cmul(temp1, Cconj(*(p+k)));
                                sad.r = sad.r + (*(error_array+k)) * temp1.r;
                                sad.i = sad.i + (*(error_array+k)) * temp1.i;
                                temp2 = Cmul( Cconj(*(p+k)) , (*(p+k)) ); 
                                temp2 = Cmul( (*(z+k)), temp2 );
                                san.r = san.r + (*(error_array+k)) * temp2.r;
                                san.i = san.i + (*(error_array+k)) * temp2.i;

                        }
                        sn.r = sn.r * total_number / error_sum;
                        sn.i = sn.i * total_number / error_sum;
                        sc.r = sc.r * total_number / error_sum;
                        sc.i = sc.i * total_number / error_sum;
                        sad.r = sad.r * total_number / error_sum;
                        sad.i = sad.i * total_number / error_sum;
                        san.r = san.r * total_number / error_sum;
                        san.i = san.i * total_number / error_sum;
                        
                        a = Cdiv(san, sad);
                        (*(c+n)) = Cdiv( sc, Csqrt(sn) ) ;

                        temp = Cmul( (*(c+n)) , Cconj(*(c+n)) );
                        F = F + temp.r;
        }

        F = ( factor1 * F ) / (factor2 * (variance - F) );

        free(c);        
        free(z);
        free(zn);
        free(p);
        free(cf);

        return F;
}
#else
long double cal_amplitude(long double *date_array, long double *norm_mag, long double *error_array, int total_number, long double trial_freq, long double variance, long double degree, long double factor1, long double factor2)
{
	complex *c, *z, *zn, *p, *cf, temp, temp1, temp2;
	long double F,sav;
	int n, k;
	complex a, sn, sc, sad, san;

	c = calloc((factor2+1), sizeof(complex));
	if(c == NULL) {fprintf(stderr, "Memory problem: c\n"); abort();}
	z = calloc(total_number, sizeof(complex));
	if(z == NULL) {fprintf(stderr, "Memory problem: z\n"); abort();}
	zn = calloc(total_number, sizeof(complex));
	if(zn == NULL) {fprintf(stderr, "Memory problem: zn\n"); abort();}
	p = calloc(total_number, sizeof(complex));
	if(p == NULL) {fprintf(stderr, "Memory problem: p\n"); abort();}
	cf = calloc(total_number, sizeof(complex));
	if(cf == NULL) {fprintf(stderr, "Memory problem: cf\n"); abort();}
	
	F = 0.0;
	a.r = 0.0;
	a.i = 0.0;

	for( n=0; n < total_number; n++) {
			sav = fmod((trial_freq * (*(date_array+n))) , (2.0 * M_PI) );
	       	(*(z+n)).r = cos(sav);
	       	(*(z+n)).i = sin(sav);
			(*(zn+n)).r = (Cconj(*(z+n))).r;
			(*(zn+n)).i = (Cconj(*(z+n))).i;
			(*(p+n)).r = (*(zn+n)).r;
			(*(p+n)).i = (*(zn+n)).i;
			(*(cf+n)).r = (*(norm_mag+n)) * cos(sav * degree);
			(*(cf+n)).i = (*(norm_mag+n)) * sin(sav * degree);
	}


	for( n=0; n <= factor2; n++) {

			sn.r = 0.0;
			sn.i = 0.0;
			sc.r = 0.0;
			sc.i = 0.0;
			sad.r = 0.0;
			sad.i = 0.0;
			san.r = 0.0;
			san.i = 0.0;

			for( k = 0; k < total_number; k++) {
				
				temp1 = Cmul((*(z+k)), (*(p+k)));
				temp2 = Cmul((*(zn+k)), Cconj(*(p+k)) );
				temp2 = Cmul(a , temp2);
				(*(p+k)).r = temp1.r - temp2.r;
				(*(p+k)).i = temp1.i - temp2.i;

				temp1 = Cmul( *(p+k) , Cconj(*(p+k)) );
				sn.r = sn.r + temp1.r;
				sn.i = sn.i + temp1.i;
				temp2 = Cmul((*(cf+k)), Cconj(*(p+k)) );
				sc.r = sc.r + temp2.r;
				sc.i = sc.i + temp2.i;

				(*(zn+k)) = Cmul((*(zn+k)), (*(z+k)));
				
				
				temp1 = Cmul((*(zn+k)), Cconj(*(p+k)));
				temp1 = Cmul(temp1, Cconj(*(p+k)));
				sad.r = sad.r + temp1.r;
				sad.i = sad.i + temp1.i;
				temp2 = Cmul( Cconj(*(p+k)) , (*(p+k)) ); 
				temp2 = Cmul( (*(z+k)), temp2 );
				san.r = san.r + temp2.r;
				san.i = san.i + temp2.i;

			}
			
			a = Cdiv(san, sad);
			(*(c+n)) = Cdiv( sc, Csqrt(sn) ) ;

			temp = Cmul( (*(c+n)) , Cconj(*(c+n)) );
			F = F + temp.r;
	}

	F = ( factor1 * F ) / (factor2 * (variance - F) );

	free(c);	
	free(z);
	free(zn);
	free(p);
	free(cf);
	
	return F;
}
#endif

complex Csqrt(complex z)
{
	complex c;
	long double x, y, w, r;

	c.r = 0.0;
	c.i = 0.0;
	
	if ((z.r == 0.0) &&  (z.i == 0.0)) {
		c.r = 0.0;
		c.i = 0.0;
		return c;
	} else {
		x = fabs(z.r);
		y = fabs(z.i);
		if (x >= y) {
			r = y/x;
			w = sqrt(x) * sqrt(0.5 * (1.0 + sqrt(1.0+r*r)));
		} else {
			r = x/y;
			w = sqrt(y) * sqrt(0.5 * (r + sqrt(1.0 + r*r)));
		}

		if (z.r >= 0.0 ) {
			c.r = w;
			c.i = z.i / (2.0*w);
		} else {
			c.i = (z.i >= 0) ? w : -w;
			c.r = y / (2.0*c.i);
		}

		return c;
	}
}

complex Cdiv(complex a, complex b)
{
	complex c;
	long double r, den;

	c.r = 0.0;
	c.i = 0.0;
	
	if (fabs(b.r) >= fabs(b.i)) {
		r = b.i / b.r;
		den = b.r + r * b.i;
		c.r = (a.r + r * a.i) / den;
		c.i = (a.i - r * a.r) / den;
	} else {
		r = b.r / b.i;
		den = b.i + r * b.r;
		c.r = (a.r * r + a.i) / den;
		c.i = (a.i * r - a.r) / den;
	}

	return c;
}
complex Cmul(complex a, complex b)
{
	complex c;

	c.r = 0.0;
	c.i = 0.0;
	
	c.r = a.r * b.r - a.i * b.i;
	c.i = a.i * b.r + a.r * b.i;

	return c;
}

complex Cconj(complex z)
{
		complex c;

		c.r = 0.0;
		c.i = 0.0;

		c.r = z.r;
		c.i = -z.i;

		return c;
}

#ifdef USE_RESAMPLE
#ifdef USE_ERR
int jackknife_sample(long double *date_array, long double *mag_array, long double *error_array, int total_number, int out_ind, long double error_sum, long double *jackknife_date_array, long double *jackknife_norm_mag, long double *jackknife_error_array, long double *new_variance, long double *sample_error_sum)
{
        int i, k, total_number_minus_1;
        long double new_mean_mag, ld_temp;

        total_number_minus_1 = total_number - 1;
        *sample_error_sum = error_sum - *(error_array + out_ind); // re-estimate the sum.
    
        k = 0;
        new_mean_mag = 0.0; // re-estimate the mean and variance.
        for(i=0; i < total_number; ++i) {
                if(i != out_ind) {
                        *(jackknife_date_array + k) = *(date_array + i); 
                        *(jackknife_norm_mag + k) = *(mag_array + i); 
                        *(jackknife_error_array + k) = *(error_array + i); 
                        k = k + 1;
                }
        }
        new_mean_mag = gsl_stats_long_double_wmean(jackknife_error_array, 1, jackknife_norm_mag, 1, total_number_minus_1);
        ld_temp = gsl_stats_long_double_wvariance_m(jackknife_error_array, 1, jackknife_norm_mag, 1, total_number_minus_1, new_mean_mag);
        *new_variance = ld_temp * (total_number_minus_1 - 1.0);
        for(k=0; k < total_number_minus_1; ++k) {
                *(jackknife_norm_mag + k) = *(jackknife_norm_mag + k) - new_mean_mag;
        }

        return 0;
}
#else
int jackknife_sample(long double *date_array, long double *mag_array, long double *error_array, int total_number, int out_ind, long double *jackknife_date_array, long double *jackknife_norm_mag, long double *jackknife_error_array, long double *new_variance)
{
        int i, k, total_number_minus_1;
        long double new_mean_mag, ld_temp;

        total_number_minus_1 = total_number - 1;
        
        k = 0;
        new_mean_mag = 0.0;  // re-estimate the mean and variance.
        for(i=0; i < total_number; ++i) {
                if(i != out_ind) {
                        *(jackknife_date_array + k) = *(date_array + i);
                        *(jackknife_norm_mag + k) = *(mag_array + i);
                        new_mean_mag = new_mean_mag + *(jackknife_norm_mag + k);
                        *(jackknife_error_array + k) = *(error_array + i);
                        k = k + 1;
                }
        }
        new_mean_mag = new_mean_mag / total_number_minus_1;

        ld_temp = 0.0;
        for(k=0; k < total_number_minus_1; ++k) {
                *(jackknife_norm_mag + k) = *(jackknife_norm_mag + k) - new_mean_mag;
                ld_temp = ld_temp + (*(jackknife_norm_mag + k)) * (*(jackknife_norm_mag + k));
        }
        *new_variance = ld_temp;

        return 0;
}
#endif
#endif
